'''
This script is based on the code from this link: https://github.com/FluVigilanciaBR/fludashboard/blob/master/Notebooks/episem.py

Brazilian epidemiological week (epiweek) is definied from Sunday to Saturday. 
Wednesday is defined as the turning point for changing from one year to another and deciding wether Jan 1st is inlcuded in the first epiweek of the new year or still in the last epiweek of the previous one.
That is, if the weekday of Jan 1st is between Sunday to Wednesday (included), then it is epiweek 01 of the new year. If the weekday is between Thursday-Saturday, then it falls in the last epiweek of the previous year (typically epiweek 52).
'''

from datetime import datetime, timedelta

__all__ = ["EpiWeek"]


class EpiWeek:
    @staticmethod
    def weekday(date):
        """
        Extract weekday from date.
        :param date: datetime.date
        :return: int [0-6] (Sun-Sat)
        """
        return int(date.strftime('%w'))

    @staticmethod
    def firstday_epiyear(year):
        """
        Get the first day of the first epidemiological week of the year.
        :param year: int
        :return: datetime.date
        """
        date = datetime(year, 1, 1)
        weekday = EpiWeek.weekday(date)

        if weekday <= 3:
            date -= timedelta(days=weekday)
        else:
            date += timedelta(days=7 - weekday)
        return date

    @staticmethod
    def lastday_epiyear(year):
        """
        Get the last day of the last epidemiological week of the year.
        :param year: int
        :return: datetime.date
        """
        return EpiWeek.firstday_epiyear(year + 1) - timedelta(days=1)

    @staticmethod
    def epiweek(date):
        """
        Get the epidemiological week from a date.
        :param date: datetime.date
        :return: dict {'year': int, 'week': int, 'epiweek': int}
        """
        year = date.year
        firstday = EpiWeek.firstday_epiyear(year)
        lastday = EpiWeek.lastday_epiyear(year)

        if date < firstday:
            year -= 1
            firstday = EpiWeek.firstday_epiyear(year)
        elif date > lastday:
            year += 1
            firstday = EpiWeek.firstday_epiyear(year)

        week = int((date - firstday).days / 7) + 1
        epiweek = year * 100 + week

        return {'year': year, 'week': week, 'epiweek': epiweek}

    @staticmethod   
    def epidate(year, week, first=True):
        """
        Get the date corresponding to the first day of the epidemiological week.
        :param year: int
        :param week: int
        :param first: bool, if True returns the first day of the week, else returns the last day
        :return: datetime.date
        """
        firstday = EpiWeek.firstday_epiyear(year)
        if first:
            return firstday + timedelta(days=(week - 1) * 7)
        return firstday + timedelta(days=week * 7 - 1)


import pandas as pd
import numpy as np
from typing import Tuple

class WeightedClassifier:
    """
    A classifier that adjusts predicted class probabilities using a weighting mechanism,
    such as a density estimation model over a continuous variable (e.g., age).

    Attributes
    ----------
    model_prob : object
        Trained classification model with `predict_proba` and `feature_name_`.
    model_density : object
        Trained density estimation model with `predict_bin_probs`.

    Methods
    -------
    predict_proba(X, range_density, num_bins):
        Returns class probabilities weighted by the distribution of a continuous variable.
    """

    def __init__(self, model_prob, model_density):
        self.model_prob = model_prob
        self.model_density = model_density

        # Extract model features
        self.cols_X = self.model_prob.model.feature_name_
        self.cols_X_density = self.model_density.model_mu.model.feature_name_

        # Identify the variable used in model_prob but not in model_density
        diff = [col for col in self.cols_X if col not in self.cols_X_density]
        if not diff:
            raise ValueError("model_prob must contain a variable not used in model_density.")
        self.var_density = diff[0]

    def predict_proba(
        self,
        X: pd.DataFrame,
        range_density: Tuple[float, float] = (0, 120),
        num_bins: int = 100
    ) -> pd.DataFrame:
        """
        Computes class probabilities weighted by the distribution of a continuous variable (e.g., age).

        Parameters
        ----------
        X : pd.DataFrame
            Input features for both models.
        range_density : tuple of float
            Range (min, max) for binning the continuous variable.
        num_bins : int
            Number of bins for the density estimation.

        Returns
        -------
        pd.DataFrame
            Aggregated class probabilities by each unique configuration in `cols_X_density`.
        """
        if not isinstance(X, pd.DataFrame):
            raise TypeError("X must be a pandas DataFrame.")
        if num_bins <= 0:
            raise ValueError("num_bins must be a positive integer.")
        if range_density[0] >= range_density[1]:
            raise ValueError("Invalid range_density: min should be less than max.")

        # Get conditioning variables (e.g., all features except age)
        X_density = X[self.cols_X_density].drop_duplicates().reset_index(drop=True)

        # Predict bin probabilities for each group
        bin_probs, centers = self.model_density.predict_bin_probs(
            X_density,
            min_val=range_density[0],
            max_val=range_density[1],
            num_bins=num_bins
        )

        bin_df = pd.DataFrame(bin_probs, columns=centers)
        X_range = X_density.join(bin_df).melt(
            id_vars=self.cols_X_density,
            var_name=self.var_density,
            value_name='VAR_PROB'
        ).astype({self.var_density: float})

        # Predict class probabilities
        probs = self.model_prob.model.predict_proba(X_range[self.cols_X])
        classes = list(self.model_prob.model.classes_)

        # Combine and weight by density
        X_range[classes] = probs * X_range[['VAR_PROB']].values

        # Aggregate to get final probabilities per subgroup
        return X_range[self.cols_X_density + classes].groupby(self.cols_X_density).sum().reset_index()



from ._gbm_trainer import GBMTrainer
from ._lognormal_estimator import LognormalEstimator
from ._weighted_classifier import WeightedClassifier

__all__ = ['GBMTrainer', 'LognormalEstimator','WeightedClassifier']


import numpy as np
from sklearn.metrics import brier_score_loss
from sklearn.model_selection import train_test_split
import lightgbm as lgb

__all__ = ['GBMTrainer']

class GBMTrainer:
    """
    GBMTrainer: A wrapper for training LightGBM models with support for regression, binary, and multiclass classification.

    This class simplifies the training process with built-in data splitting, early stopping, default hyperparameters,
    and optional refitting on the full dataset using the best number of boosting rounds.

    Attributes
    ----------
    objective : str
        Type of task to solve: 'regression', 'binary', or 'multiclass'.
    eval_metric : str
        Evaluation metric to use. If None, defaults to:
            - 'l2' for regression
            - 'binary_logloss' for binary classification
            - 'multi_logloss' for multiclass classification
    n_estimators : int
        Maximum number of boosting iterations.
    stopping_rounds : int
        Number of rounds with no improvement to trigger early stopping.
    train_size : float
        Proportion of the dataset to use for training (rest is for validation).
    log_period : int or None
        Period to log evaluation metrics during training. If None, logging is disabled.
    random_state : int
        Seed used for reproducibility.
    model : lightgbm.LGBMClassifier or lightgbm.LGBMRegressor
        The trained LightGBM model.
    best_iteration : int
        Number of boosting rounds selected by early stopping.

    Methods
    -------
    fit(X, y, verbose=False, refit=True):
        Train the model using LightGBM with early stopping on a validation split.
        If `refit=True`, retrain the model on the full dataset using the best iteration.

    brier_loss(y_true, y_pred):
        Compute the Brier score loss for probabilistic predictions.

    Examples
    --------
    >>> trainer = GBMTrainer(objective='binary')
    >>> model = trainer.fit(X_train, y_train, verbose=True, refit=True)
    >>> loss_name, score, is_higher_better = trainer.brier_loss(y_test, model.predict_proba(X_test)[:, 1])
    """

    def __init__(self, objective='multiclass', eval_metric=None
                 , n_estimators=5000, stopping_rounds=10, train_size=0.8, log_period=None, random_state=0):
        """
        Initialize the GBMTrainer.

        Parameters
        ----------
        objective : str
            Task type: 'regression', 'binary', or 'multiclass'.
        eval_metric : str or None
            Evaluation metric to use. Defaults to task-specific metrics if None.
        n_estimators : int
            Number of boosting iterations.
        stopping_rounds : int
            Number of rounds with no improvement for early stopping.
        train_size : float
            Proportion of the dataset used for training.
        log_period : int or None
            Frequency of logging during training.
        random_state : int
            Seed for reproducibility.
        """        
        self.objective = objective
        self.n_estimators = n_estimators
        self.stopping_rounds = stopping_rounds
        self.train_size = train_size
        self.log_period = log_period
        self.random_state = random_state

        valid_objectives = ['regression', 'multiclass', 'binary']
        if self.objective not in valid_objectives:
            raise ValueError(f"Objective '{self.objective}' is not supported. Choose from {valid_objectives}.")

        if eval_metric is None:
            self.eval_metric = {
                'regression': 'l2',
                'multiclass': 'multi_logloss',
                'binary': 'binary_logloss'
            }[self.objective]
        else:
            self.eval_metric = eval_metric

        self.model_class = lgb.LGBMRegressor if self.objective == 'regression' else lgb.LGBMClassifier

    def fit(self, X, y, verbose=False, refit = True):
        """
        Fit a LightGBM model with training and validation split, early stopping,
        and optional refitting on the full data using the best iteration.

        Parameters
        ----------
        X : pd.DataFrame or np.ndarray
            Feature matrix.
        y : pd.Series or np.ndarray
            Target vector.
        verbose : bool
            Whether to print logs during training.
        refit : bool
            If True, retrain model on full dataset with best number of iterations.

        Returns
        -------
        model : LGBMClassifier or LGBMRegressor
            Trained LightGBM model.
        """
        if self.model_class == lgb.LGBMClassifier:
            values, counts = np.unique(y, return_counts=True)
            remove_category = values[counts < 2]
            if len(remove_category) > 0:
                remove_index = y.isin(remove_category)
                X = X[~remove_index]
                y = y[~remove_index]
                
        stratify = y if self.model_class == lgb.LGBMClassifier else None
        X_treino, X_eval, y_treino, y_eval = train_test_split(
            X, y, train_size=self.train_size, stratify=stratify, random_state=self.random_state
        )

        self.model = self.model_class(
            boosting_type='gbdt',
            objective=self.objective,
            n_estimators=self.n_estimators,
            num_leaves=10,
            max_depth=3,
            n_jobs=-1,
            learning_rate=0.1,
            min_data_in_leaf=30,
            feature_fraction=0.8,
            min_data_in_bin=10,
            random_state=self.random_state,
            verbose=-1
        )

        eval_set = [(X_treino, y_treino), (X_eval, y_eval)]

        callbacks = []
        if self.log_period is not None:
            callbacks.append(lgb.log_evaluation(period=self.log_period))
        if self.stopping_rounds is not None:
            callbacks.append(lgb.early_stopping(stopping_rounds=self.stopping_rounds, first_metric_only=True, verbose=verbose))

        self.model.fit(X_treino, y_treino,
                       eval_set=eval_set,
                       eval_metric=self.eval_metric,
                       callbacks=callbacks)
        
        self.best_iteration = self.model.best_iteration_
        
        if refit:
            self.model = self.model_class(
                boosting_type='gbdt',
                objective=self.objective,
                n_estimators=self.best_iteration,
                num_leaves=10,
                max_depth=3,
                n_jobs=-1,
                learning_rate=0.1,
                min_data_in_leaf=30,
                feature_fraction=0.8,
                min_data_in_bin=10,
                random_state=self.random_state,
                verbose=-1
            )            
            self.model.fit(X, y)

        return self.model

    def brier_loss(self, y_true, y_pred):
        """
        Compute Brier score loss for probabilistic predictions.

        Parameters
        ----------
        y_true : array-like
            True binary labels.
        y_pred : array-like
            Predicted probabilities for the positive class.

        Returns
        -------
        metric_name : str
            The name of the metric: 'brier_score_loss'.
        score : float
            Computed Brier score.
        is_higher_better : bool
            Whether higher score indicates better performance (False for Brier loss).
        """
        is_higher_better = False
        score = brier_score_loss(y_true, y_pred)
        return "brier_score_loss", score, is_higher_better


from setuptools import setup, find_packages

setup(
    name="PySRAG",
    version="0.3.2",
    packages=find_packages(),

    # Metadata
    author="João Flávio Andrade Silva",
    author_email="joaoflavio1988@gmail.com",
    description="This Python package provides tools for analyzing and processing data related to Severe Acute Respiratory Syndrome (SARS) and other respiratory viruses. It includes functions for data preprocessing, feature engineering, and training Gradient Boosting Models (GBMs) for binary or multiclass classification.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/joao-1988/PySRAG",
    
    # License metadata
    license= "MIT", 
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],

    # Optional
    install_requires=['numpy', 'pandas', 'scikit-learn', 'lightgbm', 'datetime'], #'lightgbm==4.0.0','numpy==1.23.5','pandas==1.5.3','scikit-learn==1.2.2'
    package_data={"": ["*.csv"]}
)


import os
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
from ._epiweek import EpiWeek

__all__ = ["SRAG"]

class SRAG:
    """
    SRAG: A class for processing and preparing Sindrome Respiratória Aguda Grave (SRAG) surveillance data 
    from CSV files for machine learning tasks, especially in epidemiological modeling.

    This class reads one or more CSV files in a specific format, processes the data (cleaning, transformation,
    filtering by SRAG case definitions), and computes epidemiological features, including age normalization,
    virus positivity flags, and epidemiological week tracking.

    Attributes:
    -----------
    filepath : str or list
        Path(s) to the input CSV file(s). Filenames must include the date in DD-MM-YYYY format (e.g., 'INFLUD22-03-04-2023.csv').
    
    data : pd.DataFrame
        Processed and filtered dataset containing relevant epidemiological and demographic variables.
    
    dt_file : datetime
        Date extracted from the filename.
    
    ano_file : int
        Epidemiological year corresponding to the file date.
    
    sem_file : int
        Epidemiological week number corresponding to the file date.
    
    dt_file_sem : datetime
        Date corresponding to the Sunday of the epidemiological week of the file date.

    Methods:
    --------
    generate_training_data(objective, cols_X, col_y):
        Generates the training dataset (X, y) for machine learning. Supports both binary/multiclass/regression tasks.

    generate_training_weeks():
        Returns a DataFrame listing all epidemiological weeks present in the processed dataset.

    get_start_day_of_week(lag=0):
        Returns a dictionary with information about the start day of the epidemiological week, considering an optional lag.

    load_common_data():
        Loads external metadata (e.g., geolocation, population) for Brazilian municipalities from 'common_data.csv'.

    Example:
    --------
    >>> srag = SRAG("INFLUD22-03-04-2023.csv")
    >>> X, y = srag.generate_training_data(objective="multiclass", cols_X=["LATITUDE", "LONGITUDE", "IDADE_ANO"], col_y=["POS_SARS2", "POS_FLUA", "POS_FLUB", "POS_VSR"])
    >>> srag.data.head()
    """

    def __init__(self, filepath, col_type_add={}, col_out_add=[]):
        """
        Initialize SRAG object with file path(s) and optional custom columns.

        Parameters:
        -----------
        filepath : str or list of str
            Path(s) to CSV file(s). Must include date in DD-MM-YYYY format.
        col_type_add : dict, optional
            Dictionary of additional columns and their types to load.
        col_out_add : list, optional
            List of additional output columns to retain after processing.
        """
        self.__filepath = filepath
        if isinstance(self.__filepath, list):
            self.__dt_file = max([datetime.strptime(fp[-14:-4], '%d-%m-%Y') for fp in self.__filepath])
        else:
            self.__dt_file = datetime.strptime(self.__filepath[-14:-4], '%d-%m-%Y')   
        self.__ano_file = EpiWeek.epiweek(self.__dt_file)['year']
        self.__sem_file = EpiWeek.epiweek(self.__dt_file)['week']
        self.__dt_file_sem = self.__get_previous_sunday(self.__dt_file)
        self.__load_file_data(col_type_add, col_out_add)
        self.__all_viruses = ['POS_FLUA', 'POS_FLUB', 'POS_SARS2', 'POS_VSR', 
                              'POS_PARA1','POS_PARA2', 'POS_PARA3', 'POS_PARA4',
                              'POS_ADENO', 'POS_METAP','POS_BOCA', 'POS_RINO', 'POS_OUTROS']

    def __get_previous_sunday(self, date):
        days_until_sunday = int(date.strftime('%w'))
        previous_sunday = date - timedelta(days=days_until_sunday)
        return previous_sunday

    def __process_data(self, filepath, col_type_add = {}, col_out_add = []):
        
        col_type = {'DT_NOTIFIC': str
            , 'DT_DIGITA': str
            , 'DT_SIN_PRI': str
            , 'SEM_PRI': str
            , 'SG_UF_NOT': str
            , 'ID_MUNICIP': str
            , 'CO_MUN_NOT': str
            , 'CS_SEXO': str
            , 'NU_IDADE_N': float
            , 'TP_IDADE': float
            , 'AMOSTRA': float
            , 'PCR_RESUL': float
            , 'RES_AN': float
            , 'POS_PCRFLU': float
            , 'TP_FLU_PCR': float
            , 'POS_AN_FLU': float
            , 'TP_FLU_AN': float
            , 'POS_PCROUT': float
            , 'PCR_SARS2': float
            , 'POS_AN_OUT': float
            , 'AN_SARS2': float
            , 'PCR_VSR': float
            , 'AN_VSR': float
            , 'PCR_PARA1': float
            , 'AN_PARA1': float
            , 'PCR_PARA2': float
            , 'AN_PARA2': float
            , 'PCR_PARA3': float
            , 'AN_PARA3': float
            , 'PCR_PARA4': float
            , 'PCR_ADENO': float
            , 'AN_ADENO': float
            , 'PCR_METAP': float
            , 'PCR_BOCA': float
            , 'PCR_RINO': float
            , 'PCR_OUTRO': float
            , 'AN_OUTRO': float
            , 'HOSPITAL': float
            , 'EVOLUCAO': str
            , 'TOSSE': float
            , 'FEBRE': float
            , 'GARGANTA': float
            , 'DISPNEIA': float
            , 'DESC_RESP': float
            , 'SATURACAO': float
            }
        
        col_out = ['DT_FILE', 'ANO_FILE', 'SEM_FILE', 'DT_FILE_SEM'
                , 'DT_SIN_PRI', 'ANO_SIN_PRI', 'SEM_SIN_PRI', 'ANO_SEM_SIN_PRI'
                , 'DT_SIN_PRI_SEM', 'DIF_SEM_FILE_SIN_PRI'
                , 'REGIAO', 'REGIAO_LATITUDE', 'REGIAO_LONGITUDE'
                , 'SG_UF_NOT', 'UF_LATITUDE', 'UF_LONGITUDE'
                , 'ID_MUNICIP', 'CO_MUN_NOT', 'LATITUDE', 'LONGITUDE'
                , 'POPULACAO'
                , 'IDADE_ANO', 'CS_SEXO'
                , 'AMOSTRA', 'PCR_RESUL', 'RES_AN'
                , 'POS_FLUA', 'POS_FLUB', 'POS_SARS2', 'POS_VSR'
                , 'POS_PARA1', 'POS_PARA2', 'POS_PARA3', 'POS_PARA4'
                , 'POS_ADENO', 'POS_METAP', 'POS_BOCA'
                , 'POS_RINO', 'POS_OUTROS'
                , 'POS_SUM', 'POS_ANY'
                ]
        
        col_type.update(col_type_add)
        col_out_add = list(set(col_out_add) - (set(col_out)))
        col_out.extend(col_out_add)
        
        data = (pd.read_csv(filepath, sep=';', encoding='latin-1', engine='pyarrow'
                            , usecols=col_type.keys(), dtype=col_type)
                .assign( evolucao = lambda x: pd.to_numeric(x['EVOLUCAO'], errors='coerce')
                        ,sin_SG = lambda x: np.where( (x['TOSSE'] == 1) | (x['GARGANTA'] == 1), 1, 0)
                        ,sin_ad_SRAG = lambda x: np.where( (x['DISPNEIA'] == 1) | (x['DESC_RESP'] == 1) | (x['SATURACAO'] == 1), 1, 0)
                        ,cond_SRAG = lambda x: np.where( ( (x['HOSPITAL'] == 1) | (x['evolucao'] == 2) ) &
                                                        (x['sin_SG'] == 1) & (x['sin_ad_SRAG'] == 1), 1, 0)
                        )
                .query('cond_SRAG == 1')
                )

        formats_to_try = ['%d/%m/%Y', '%Y-%m-%d']
        DT_SIN_PRI = data['DT_SIN_PRI']
        for fmt in formats_to_try:
            try:
                data['DT_SIN_PRI'] = pd.to_datetime(DT_SIN_PRI, format=fmt)
                break
            except ValueError:
                continue

        data_processed = (
            data
            .assign(
                DT_FILE=self.__dt_file
                , ANO_FILE=self.__ano_file
                , SEM_FILE=self.__sem_file
                , DT_FILE_SEM=self.__dt_file_sem

                , ANO_SIN_PRI = lambda x: x['DT_SIN_PRI'].apply(lambda x: EpiWeek.epiweek(x)['year'])            
                , SEM_SIN_PRI = lambda x: x['DT_SIN_PRI'].apply(lambda x: EpiWeek.epiweek(x)['week']) 
                , ANO_SEM_SIN_PRI=lambda x: x['DT_SIN_PRI'].apply(lambda x: EpiWeek.epiweek(x)['epiweek']) 

                , DT_SIN_PRI_SEM=lambda x: x['DT_SIN_PRI'].apply(self.__get_previous_sunday)
                , DIF_SEM_FILE_SIN_PRI=lambda x: ((x['DT_FILE_SEM'] - x['DT_SIN_PRI_SEM']).dt.days / 7).astype(int)

                , IDADE_ANO_aux=lambda x: np.where(x['TP_IDADE'] == 3, np.round(x['NU_IDADE_N'], 4),
                                    np.where(x['TP_IDADE'] == 2, np.round(x['NU_IDADE_N'] /  (12), 4),
                                    np.where(x['TP_IDADE'] == 1, np.round(x['NU_IDADE_N'] / (365), 4), None))).astype(float)
                , IDADE_ANO=lambda x: np.where(x['IDADE_ANO_aux'] <= 0, np.round(1/365, 4) , x['IDADE_ANO_aux'])
                , CO_MUN_NOT=lambda x: (x['CO_MUN_NOT']).astype(float).fillna(0).astype(int)
                , POS_FLUA=lambda x: np.where(((x['POS_PCRFLU'] == 1) & (x['TP_FLU_PCR'] == 1)) |
                                            ((x['POS_AN_FLU'] == 1) & (x['TP_FLU_AN'] == 1))
                                            , 1, 0)
                , POS_FLUB=lambda x: np.where(((x['POS_PCRFLU'] == 1) & (x['TP_FLU_PCR'] == 2)) |
                                            ((x['POS_AN_FLU'] == 1) & (x['TP_FLU_AN'] == 2))
                                            , 1, 0)
                , POS_SARS2=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_SARS2'] == 1)) |
                                            ((x['POS_AN_OUT'] == 1) & (x['AN_SARS2'] == 1))
                                            , 1, 0)
                , POS_VSR=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_VSR'] == 1)) |
                                            ((x['POS_AN_OUT'] == 1) & (x['AN_VSR'] == 1))
                                            , 1, 0)
                , POS_PARA1=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_PARA1'] == 1)) |
                                            ((x['POS_AN_OUT'] == 1) & (x['AN_PARA1'] == 1))
                                            , 1, 0)
                , POS_PARA2=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_PARA2'] == 1)) |
                                            ((x['POS_AN_OUT'] == 1) & (x['AN_PARA2'] == 1))
                                            , 1, 0)
                , POS_PARA3=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_PARA3'] == 1)) |
                                            ((x['POS_AN_OUT'] == 1) & (x['AN_PARA3'] == 1))
                                            , 1, 0)
                , POS_PARA4=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_PARA4'] == 1))
                                            , 1, 0)
                , POS_ADENO=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_ADENO'] == 1)) |
                                            ((x['POS_AN_OUT'] == 1) & (x['AN_ADENO'] == 1))
                                            , 1, 0)
                , POS_METAP=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_METAP'] == 1))
                                            , 1, 0)
                , POS_BOCA=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_BOCA'] == 1))
                                            , 1, 0)
                , POS_RINO=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_RINO'] == 1))
                                            , 1, 0)
                , POS_OUTROS=lambda x: np.where(((x['POS_PCROUT'] == 1) & (x['PCR_OUTRO'] == 1)) |
                                                ((x['POS_AN_OUT'] == 1) & (x['AN_OUTRO'] == 1))
                                                , 1, 0)
                , POS_SUM=lambda x: x[['POS_FLUA', 'POS_FLUB', 'POS_SARS2', 'POS_VSR', 'POS_PARA1',
                                    'POS_PARA2', 'POS_PARA3', 'POS_PARA4', 'POS_ADENO', 'POS_METAP',
                                    'POS_BOCA', 'POS_RINO', 'POS_OUTROS']].sum(axis=1)
                , POS_ANY=lambda x: np.where(x['POS_SUM'] > 0, 1, 0)
                )
                .merge(self.load_common_data(), how='left', left_on='CO_MUN_NOT', right_on='CD_IBGE')
                [col_out]
                .sort_values('DT_SIN_PRI', ascending=False)
                .reset_index(drop=True)
                )
        return data_processed
        
    def __load_file_data(self, col_type_add = {}, col_out_add = []):

        if isinstance(self.__filepath, list):
            data = pd.concat([ self.__process_data(fp, col_type_add, col_out_add) for fp in self.__filepath ]).reset_index(drop=True)
        else:
            data = self.__process_data(self.__filepath, col_type_add, col_out_add)

        self.__data = data

    def __dummy_to_label(self, srs):
        return np.where(srs == 1, srs.name[4:], None)

    def __remove_none(self, lst):
        return [i for i in lst if i != None]

    def generate_training_data(self, objective, cols_X, col_y):  
        """
        Generate training data from SRAG records based on a specific objective.

        Parameters:
        -----------
        objective : str
            Either 'multiclass' or another objective (treated as binary or regression).
        cols_X : list
            List of column names to use as input features.
        col_y : str or list
            Target variable(s). For multiclass, a list of virus labels is expected.

        Returns:
        --------
        X : pd.DataFrame
            Features dataframe.
        y : pd.Series
            Target variable or virus label.
        """             
        df = self.__data.query('POS_ANY > 0')

        if objective == 'multiclass':           
            residual_viruses = [virus for virus in self.__all_viruses if virus not in col_y]
            if len(residual_viruses) > 0:
                df = df.assign(POS_RESIDUAL=lambda x: x[residual_viruses].max(axis=1))
                col_y.extend(['POS_RESIDUAL'])

            virus = (df[col_y]
                     .apply(self.__dummy_to_label)
                     .apply(self.__remove_none, axis=1))
            df = (df[cols_X]
                  .assign(VIRUS=virus)
                  .explode('VIRUS')
                  .reset_index(drop=True))
            X, y = df[cols_X], df['VIRUS']
        else:
            df = df.reset_index(drop=True)
            X, y = df[cols_X], df[col_y]
        return X, y

    def generate_training_weeks(self):
        """
        Extracts unique weeks present in the dataset, sorted by time.

        Returns:
        --------
        pd.DataFrame
            A DataFrame with columns for week/year and their starting Sundays.
        """        
        df = (self.__data[['DT_SIN_PRI_SEM', 'ANO_SIN_PRI', 'SEM_SIN_PRI', 'ANO_SEM_SIN_PRI']]
                         .drop_duplicates()
                         .sort_values(by='ANO_SEM_SIN_PRI')
                         .reset_index(drop=True)
                         )        
        return df

    def get_start_day_of_week(self, lag=0):
        """
        Get metadata about the start date of an epidemiological week with optional lag.

        Parameters:
        -----------
        lag : int, optional
            Number of weeks before the current file week.

        Returns:
        --------
        dict
            Dictionary with keys: 'lag', 'year', 'week', 'start_day_week'.
        """        
        start_day_week = self.__dt_file_sem - timedelta(weeks=lag)
        year = EpiWeek.epiweek(start_day_week)['year']
        week = EpiWeek.epiweek(start_day_week)['week']
        return {'lag': lag, 'year': year, 'week': week, 'start_day_week': start_day_week}

    @property
    def filepath(self):
        """Returns the input file path(s)."""
        return self.__filepath

    @property
    def data(self):
        """Returns the processed SRAG dataset."""
        return self.__data

    @property
    def dt_file(self):
        """Returns the datetime object for the file date."""
        return self.__dt_file

    @property
    def ano_file(self):
        """Returns the epidemiological year of the file."""
        return self.__ano_file

    @property
    def sem_file(self):
        """Returns the epidemiological week of the file."""
        return self.__sem_file

    @property
    def dt_file_sem(self):
        """Returns the Sunday of the epidemiological week for the file date."""
        return self.__dt_file_sem
    
    @staticmethod
    def load_common_data():
        """
        Load external reference data for municipalities (e.g., location, population).

        Returns:
        --------
        pd.DataFrame
            DataFrame containing metadata from 'common_data.csv'.
        """        
        data_path = os.path.abspath(os.path.dirname(__file__))
        data_filepath = os.path.join(data_path, 'common_data.csv')
        return pd.read_csv(data_filepath)


import numpy as np
from typing import Dict, List
from scipy.stats import norm
from ._gbm_trainer import GBMTrainer

__all__ = ['LognormalEstimator']

class LognormalEstimator:
    """
    LognormalEstimator: A two-step gradient boosting model to estimate the parameters
    of a lognormal distribution (μ and σ) from tabular data.

    This estimator is useful for modeling strictly positive, skewed response variables
    under the assumption that the logarithm of the response is approximately normally 
    distributed conditional on the input features.

    Attributes
    ----------
    objective : str
        Task type, fixed to 'regression'.
    eval_metric : str
        Evaluation metric for LightGBM training, fixed to 'l2' loss.
    n_estimators : int
        Maximum number of boosting iterations.
    stopping_rounds : int
        Number of rounds with no improvement to trigger early stopping.
    train_size : float
        Proportion of data used for training (rest for validation).
    log_period : int or None
        Frequency of logging evaluation metrics. If None, logging is disabled.
    random_state : int
        Seed for reproducibility.
    model_mu : GBMTrainer
        Model trained to predict the mean (μ) of the log-transformed target.
    model_logsigma2 : GBMTrainer
        Model trained to predict the log of the variance of the log-transformed target.

    Methods
    -------
    fit(X, y, verbose=False, refit=True):
        Trains two LightGBM models to estimate μ and σ of a lognormal distribution from input features.

    predict_parameters(X):
        Returns estimated μ and σ from the trained models for given input features.

    predict_bin_probs(X, min_val, max_val, num_bins):
        Predicts probability distribution over user-defined value bins for each sample in X.

    get_bin_probs_batch(mu, sigma, min_val, max_val, num_bins):
        Vectorized version of bin probability estimation given arrays of μ and σ.

    _generate_bins(min_val, max_val, num_bins):
        Generates bin edges and their logarithmic transformation for use in PDF/CDF calculations.

    _get_bin_probs(mu, sigma, bins):
        Computes probability mass assigned to each bin by the lognormal distribution parameterized by μ and σ.

    Example
    -------
    >>> estimator = LognormalEstimator()
    >>> estimator.fit(X_train, y_train)
    >>> mu, sigma = estimator.predict_parameters(X_test)
    >>> probs, centers = estimator.predict_bin_probs(X_test, min_val=1, max_val=100, num_bins=10)
    """

    def __init__(
        self,
        n_estimators: int = 10000,
        stopping_rounds: int = 10,
        train_size: float = 0.8,
        log_period: int = None,
        random_state: int = 0
    ):
        self.objective = 'regression'
        self.eval_metric = 'l2'
        self.n_estimators = n_estimators
        self.stopping_rounds = stopping_rounds
        self.train_size = train_size
        self.log_period = log_period
        self.random_state = random_state

    def fit(self, X, y, verbose: bool = False, refit: bool = True):
        """
        Fits lognormal distribution parameters using gradient boosting models on log-transformed data.

        Parameters
        ----------
        X : array-like
            Feature matrix.
        y : array-like
            Target values (must be strictly positive).
        verbose : bool, optional
            Whether to log LightGBM training output.
        refit : bool, optional
            If True, refit the models on the entire dataset using best iteration.
        """        
        if np.min(y) <= 0:
            raise ValueError("Target values `y` must be positive for lognormal fitting.")

        # Fit model for μ (mean of log(y))
        log_y = np.log(y)
        self.model_mu = GBMTrainer(
            objective=self.objective,
            eval_metric=self.eval_metric,
            n_estimators=self.n_estimators,
            stopping_rounds=self.stopping_rounds,
            train_size=self.train_size,
            log_period=self.log_period,
            random_state=self.random_state
        )
        self.model_mu.fit(X, log_y, verbose=verbose, refit=refit)

        # Fit model for log(σ²)
        pred_log_y = self.model_mu.model.predict(X)
        log_sigma2 = np.log((log_y - pred_log_y) ** 2)
        self.model_logsigma2 = GBMTrainer(
            objective=self.objective,
            eval_metric=self.eval_metric,
            n_estimators=self.n_estimators,
            stopping_rounds=self.stopping_rounds,
            train_size=self.train_size,
            log_period=self.log_period,
            random_state=self.random_state
        )
        self.model_logsigma2.fit(X, log_sigma2, verbose=verbose, refit=refit)    

    def predict_parameters(self, X):
        """
        Predicts parameters (μ, σ) of the lognormal distribution for each row in X.

        Parameters
        ----------
        X : array-like
            Feature matrix.

        Returns
        -------
        mu : np.ndarray
            Predicted mean of log(y).
        sigma : np.ndarray
            Predicted standard deviation of log(y).
        """
        pred_mu = self.model_mu.model.predict(X)
        pred_log_sigma2 = self.model_logsigma2.model.predict(X)
        pred_sigma = np.sqrt(np.exp(pred_log_sigma2))
        return pred_mu, pred_sigma  

    def _generate_bins(self, min_val: float, max_val: float, num_bins: int) -> Dict[str, List[float]]:
        """
        Generates bins and their logarithmic equivalents for computing probabilities.

        Parameters
        ----------
        min_val : float
            Minimum bin value (> 0).
        max_val : float
            Maximum bin value.
        num_bins : int
            Number of bins.

        Returns
        -------
        bins : dict
            Dictionary with keys: 'bin_lower', 'bin_upper', 'bin_center',
            'log_bin_lower', 'log_bin_upper', 'log_bin_center'.
        """
        if min_val < 0 or max_val <= 0:
            raise ValueError("Both min_val and max_val must be positive.")
        if max_val <= min_val:
            raise ValueError("max_val must be greater than min_val.")

        edges = np.linspace(min_val, max_val, num_bins + 1)
        bin_lower = edges[:-1]
        bin_upper = edges[1:]
        bin_center = (bin_lower + bin_upper) / 2

        # Log-transform, handling log(0)
        log_bin_lower = np.full_like(bin_lower, -np.inf)
        mask = bin_lower > 0
        log_bin_lower[mask] = np.log(bin_lower[mask])
        log_bin_upper = np.log(bin_upper)
        log_bin_center = np.log(bin_center)

        return {
            "bin_lower": bin_lower,
            "bin_upper": bin_upper,
            "bin_center": bin_center,
            "log_bin_lower": log_bin_lower,
            "log_bin_upper": log_bin_upper,
            "log_bin_center": log_bin_center
        }

    def _get_bin_probs(self, mu: float, sigma: float, bins: Dict[str, np.ndarray]) -> List[float]:
        """
        Computes the probability that a lognormal variable with given μ and σ falls into each bin.

        Parameters
        ----------
        mu : float
            Mean of the log-transformed variable.
        sigma : float
            Standard deviation of the log-transformed variable.
        bins : dict
            Dictionary of bin boundaries in both normal and log space.

        Returns
        -------
        probs : list
            List of probabilities (summing to 1) for each bin.
        """        
        log_lower = bins['log_bin_lower']
        log_upper = bins['log_bin_upper']
        bin_center = bins['bin_center']

        cdf_lower = np.zeros_like(log_lower)
        mask = log_lower > -np.inf
        cdf_lower[mask] = norm.cdf(log_lower[mask], mu, sigma)
        cdf_upper = norm.cdf(log_upper, mu, sigma)

        cdf_min = np.min(cdf_lower)
        cdf_max = np.max(cdf_upper)
        num_bins = len(bin_center)

        if cdf_max == cdf_min:
            return [1.0 / num_bins] * num_bins

        probs = (cdf_upper - cdf_lower) / (cdf_max - cdf_min)
        return probs.tolist()
    
    def get_bin_probs_batch(self, mu: np.ndarray, sigma: np.ndarray, min_val: float, max_val: float, num_bins: int):
        """
        Vectorized bin probability estimation given arrays of μ and σ.

        Parameters
        ----------
        mu : np.ndarray
            Mean values of log(y).
        sigma : np.ndarray
            Standard deviations of log(y).
        min_val : float
            Minimum value for binning (must be > 0).
        max_val : float
            Maximum value for binning.
        num_bins : int
            Number of bins.

        Returns
        -------
        probs : np.ndarray
            Matrix of shape (n_samples, num_bins) with bin probabilities.
        centers : np.ndarray
            Bin center values in original scale.
        """
        bins = self._generate_bins(min_val, max_val, num_bins)
        probs = [self._get_bin_probs(mu[i], sigma[i], bins) for i in range(len(mu))]
        return np.array(probs), np.array(bins['bin_center'])

    def predict_bin_probs(self, X, min_val: float, max_val: float, num_bins: int):
        """
        Predicts probabilities for each sample falling into user-defined value bins.

        Parameters
        ----------
        X : array-like
            Feature matrix.
        min_val : float
            Minimum value for bin range (must be > 0).
        max_val : float
            Maximum value for bin range.
        num_bins : int
            Number of bins.

        Returns
        -------
        probs : np.ndarray
            Array of shape (n_samples, num_bins) with probabilities.
        centers : np.ndarray
            Bin center values in original (non-log) scale.
        """
        mu, sigma = self.predict_parameters(X)
        return self.get_bin_probs_batch(mu, sigma, min_val, max_val, num_bins)